@@ -29,19 +29,25 @@
 MenuWidget::MenuWidget(Plasma::Applet* applet)
 : QGraphicsWidget(applet)
 , mApplet(applet)
-, mLayout(new QGraphicsLinearLayout(this))
 , mMouseChecker(new QTimer(this))
 , mUpdateButtonsTimer(new QTimer(this))
 , mRootMenu(0)
+, mOverflowButton(createButton())
 , mCurrentButton(0)
 {
+    mOverflowButton->setText(QString::fromUtf8("…"));
+    mOverflowButton->setMenu(new QMenu);
+    connect(mOverflowButton->menu(), SIGNAL(aboutToHide()), SLOT(slotAboutToHideMenu()));
+    mOverflowButton->menu()->installEventFilter(this);
+
     mMouseChecker->setInterval(100);
     connect(mMouseChecker, SIGNAL(timeout()), SLOT(checkMousePosition()));
 
     mUpdateButtonsTimer->setSingleShot(true);
     connect(mUpdateButtonsTimer, SIGNAL(timeout()), SLOT(updateButtons()));
 
-    mLayout->setContentsMargins(0, 0, 0, 0);
+    setSizePolicy(QSizePolicy::Expanding, QSizePolicy::Minimum);
+    setMinimumSize(mOverflowButton->minimumSize());
 }
 
 void MenuWidget::setMenu(QMenu* menu)
@@ -55,17 +61,14 @@ void MenuWidget::updateButtons()
 {
     mUpdateButtonsTimer->stop();
 
-    // If we must show less buttons than before, drop the last ones
-    int dropCount = mMenuButtonList.count() - mRootMenu->actions().count();
-    for (;dropCount > 0; --dropCount) {
-        delete mMenuButtonList.takeLast();
-    }
-
     // Create buttons for each actions in mRootMenu, reusing existing buttons if possible
     QList<MenuButton*>::Iterator
         it = mMenuButtonList.begin(),
         end = mMenuButtonList.end();
     Q_FOREACH(QAction* action, mRootMenu->actions()) {
+        if (!action->isVisible()) {
+            continue;
+        }
         if (action->isSeparator()) {
             continue;
         }
@@ -79,6 +82,7 @@ void MenuWidget::updateButtons()
         MenuButton* button;
         if (it == end) {
             button = createButton();
+            mMenuButtonList << button;
         } else {
             button = *it;
             ++it;
@@ -89,8 +93,16 @@ void MenuWidget::updateButtons()
         button->setMenu(menu);
         disconnect(menu, 0, this, 0); // Ensure we are only connected to aboutToHide() once
         connect(menu, SIGNAL(aboutToHide()), SLOT(slotAboutToHideMenu()));
-        action->menu()->installEventFilter(this);
+        menu->installEventFilter(this);
     }
+
+    // Delete unused buttons
+    int count = end - it;
+    for (;count > 0; --count) {
+        delete mMenuButtonList.takeLast();
+    }
+    updateButtonsGeometries();
+    updateGeometry();
 }
 
 bool MenuWidget::eventFilter(QObject* object, QEvent* event)
@@ -146,7 +158,6 @@ MenuButton* MenuWidget::createButton()
 {
     MenuButton* button = new MenuButton(this);
     button->setSizePolicy(QSizePolicy::Fixed, QSizePolicy::Minimum);
-    connect(button, SIGNAL(clicked()), SLOT(slotButtonClicked()));
 
     QGraphicsDropShadowEffect* shadow = new QGraphicsDropShadowEffect();
     shadow->setBlurRadius(5);
@@ -154,8 +165,7 @@ MenuButton* MenuWidget::createButton()
     shadow->setColor(Plasma::Theme::defaultTheme()->color(Plasma::Theme::BackgroundColor));
     setGraphicsEffect(shadow);
 
-    mLayout->addItem(button);
-    mMenuButtonList << button;
+    connect(button, SIGNAL(clicked()), SLOT(slotButtonClicked()));
     return button;
 }
 
@@ -171,17 +181,40 @@ void MenuWidget::slotButtonClicked()
 
 void MenuWidget::showNextPrevMenu(bool next)
 {
-    int index = mMenuButtonList.indexOf(mCurrentButton);
-    if (index == -1) {
-        kWarning() << "Couldn't find button!";
-        return;
-    }
-    if (next) {
-        index = (index + 1) % mMenuButtonList.count();
+    int index;
+    MenuButton* button = 0;
+    if (mCurrentButton == mOverflowButton) {
+        if (next) {
+            button = mMenuButtonList.first();
+        } else {
+            Q_FOREACH(MenuButton* tmp, mMenuButtonList) {
+                if (tmp->isVisible()) {
+                    button = tmp;
+                } else {
+                    break;
+                }
+            }
+        }
     } else {
-        index = (index == 0 ? mMenuButtonList.count() : index) - 1;
+        index = mMenuButtonList.indexOf(mCurrentButton);
+        if (index == -1) {
+            kWarning() << "Couldn't find button!";
+            return;
+        }
+        if (next) {
+            index = (index + 1) % mMenuButtonList.count();
+        } else {
+            index = (index == 0 ? mMenuButtonList.count() : index) - 1;
+        }
+        button = mMenuButtonList.at(index);
+        if (!button->isVisible()) {
+            button = mOverflowButton;
+        }
+    }
+    if (button) {
+        kWarning() << "nextPrevButton" << button->text();
+        showMenu(button);
     }
-    showMenu(mMenuButtonList.at(index));
 }
 
 void MenuWidget::showMenu(MenuButton* button)
@@ -194,6 +227,9 @@ void MenuWidget::showMenu(MenuButton* button)
         }
     }
     QMenu* menu = button->menu();
+    if (!menu) {
+        return;
+    }
     QPoint pos = mApplet->containment()->corona()->popupPosition(button, menu->sizeHint());
 
     mCurrentButton = button;
@@ -237,19 +273,14 @@ void MenuWidget::slotAboutToHideMenu()
         mCurrentButton->nativeWidget()->setDown(false);
         mCurrentButton = 0;
         mMouseChecker->stop();
-    } else {
+    } else if (mCurrentButton != mOverflowButton) {
         kWarning() << "Not called from menu associated with mCurrentButton!";
     }
 }
 
 void MenuWidget::activate()
 {
-    QGraphicsLayoutItem* layoutItem = mLayout->itemAt(0);
-    if (!layoutItem) {
-        kWarning() << "No layoutItem found!";
-        return;
-    }
-    MenuButton* button = qobject_cast<MenuButton*>(layoutItem->graphicsItem()->toGraphicsObject());
+    MenuButton* button = mMenuButtonList.first();
     if (!button) {
         kWarning() << "No button found!";
         return;
@@ -278,3 +309,76 @@ void MenuWidget::activateActionInMenu(QAction* action)
     button->nativeWidget()->animateClick();
     button->menu()->setActiveAction(action);
 }
+
+void MenuWidget::resizeEvent(QGraphicsSceneResizeEvent* event)
+{
+    QGraphicsWidget::resizeEvent(event);
+    updateButtonsGeometries();
+}
+
+void MenuWidget::updateButtonsGeometries()
+{
+    qreal x = 0;
+    bool overflow = false;
+    int widgetWidth = size().width();
+    int widgetHeight = size().height();
+    QMenu* overflowMenu = mOverflowButton->menu();
+    overflowMenu->clear();
+
+    int lastVisibleIndex = -1;
+    Q_FOREACH(MenuButton* button, mMenuButtonList) {
+        // Note: we define geometry for both visible and invisible buttons because
+        // the geometry of invisible buttons is used for sizeHint
+        qreal buttonWidth = button->preferredWidth();
+        button->setGeometry(QRectF(x, 0, buttonWidth, widgetHeight));
+        x += buttonWidth;
+        if (x <= widgetWidth) {
+            button->show();
+            ++lastVisibleIndex;
+        } else {
+            button->hide();
+            overflowMenu->addMenu(button->menu());
+            overflow = true;
+        }
+    }
+
+    if (overflow) {
+        mOverflowButton->show();
+        qreal buttonWidth = mOverflowButton->preferredWidth();
+
+        // Continue hiding buttons from the right until there is enough room for
+        // the overflow button (we use a loop in the unlikely event that more
+        // than one button need to be hidden)
+        MenuButton* lastVisibleButton = 0;
+        for (; lastVisibleIndex >= 0; --lastVisibleIndex) {
+            lastVisibleButton = mMenuButtonList.at(lastVisibleIndex);
+            if (lastVisibleButton->geometry().right() > widgetWidth - buttonWidth) {
+                lastVisibleButton->hide();
+                overflowMenu->insertMenu(overflowMenu->actions().first(), lastVisibleButton->menu());
+            } else {
+                break;
+            }
+        }
+        x = lastVisibleIndex >= 0 ? lastVisibleButton->geometry().right() : 0;
+        mOverflowButton->setGeometry(QRectF(x, 0, buttonWidth, widgetHeight));
+    } else {
+        mOverflowButton->hide();
+    }
+}
+
+QSizeF MenuWidget::sizeHint(Qt::SizeHint which, const QSizeF& constraint) const
+{
+    QSizeF fullSize(0, 0);
+    Q_FOREACH(MenuButton* button, mMenuButtonList) {
+        fullSize.rwidth() += button->minimumWidth();
+        fullSize.setHeight(qMax(fullSize.height(), button->minimumHeight()));
+    }
+    
+    if (which == Qt::MinimumSize) {
+        return mOverflowButton->minimumSize();
+    } if (which == Qt::PreferredSize) {
+        return fullSize;
+    } else {
+        return QSize(fullSize.width(), constraint.height());
+    }
+}
\ No newline at end of file
