@@ -20,7 +20,6 @@
 #include <Plasma/Applet>
 #include <Plasma/ToolButton>
 
-class QGraphicsLinearLayout;
 class QMenu;
 
 class MenuButton : public Plasma::ToolButton
@@ -67,8 +66,11 @@ public:
     void activateAction(QAction*);
     void activateActionInMenu(QAction* action);
 
+    virtual QSizeF sizeHint(Qt::SizeHint which, const QSizeF& constraint = QSize()) const;
+
 protected:
     virtual bool eventFilter(QObject*, QEvent*);
+    virtual void resizeEvent(QGraphicsSceneResizeEvent* event);
 
 private Q_SLOTS:
     void slotButtonClicked();
@@ -79,11 +81,11 @@ private Q_SLOTS:
 
 private:
     Plasma::Applet* mApplet;
-    QGraphicsLinearLayout* mLayout;
     QTimer* mMouseChecker;
     QTimer* mUpdateButtonsTimer;
     QMenu* mRootMenu;
     QList<MenuButton*> mMenuButtonList;
+    MenuButton* mOverflowButton;
     MenuButton* mCurrentButton;
 
     void showNextPrevMenu(bool next);
@@ -91,5 +93,7 @@ private:
 
     bool rootMenuEventFilter(QEvent*);
     bool subMenuEventFilter(QMenu*, QEvent*);
+
+    void updateButtonsGeometries();
 };
 #endif /* MENUWIDGET_H */
